#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.annotate
# ---------------------------------------------------------------------
# Copyright (c) 2013-2017 Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under
# the terms of the LICENCE attached in the distribution package.
#
# Created on 2012-04-03

'''Provides Python 3k forward-compatible (`3107`:pep:) annotations.'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from re import compile as _regex_compile
from ast import parse as _ast_parse

from xoutil.future.functools import partial
_ast_parse = partial(_ast_parse, filename="<annotations>", mode="eval")

from xoutil.decorator.meta import decorator

__all__ = ['annotate']


_SIGNATURE = _regex_compile(r'''(?ixm)
                            \(              # Required opening for the argumens
                            (?P<args>(.)*)
                            \)\s*           # Required close
                            (?:->\s*(?P<return>.+))?$
                            ''')

_ARG_SEP = _regex_compile(r'(?im)^\*{0,2}(?P<argname>[_\w\d]+)\s*:')


def _split_signature(signature):
    from xoutil.eight import string_types as strs
    signature = (signature.strip() if isinstance(signature, strs) else '')
    if signature:
        matches = _SIGNATURE.match(signature)
        return matches.group('args'), matches.group('return')
    else:
        return '', None


def _parse_signature(signature):
    def _split_annotation_expression(expr):
        match = _ARG_SEP.match(expr)
        if not match:
            raise SyntaxError('Invalid signature expression %r' % expr)
        argname = match.group('argname')
        expr = expr[match.end():].lstrip()
        if not argname:
            raise SyntaxError('The signature %r is not valid' % expr)
        try:
            # This is a hack to help not implement an expression parser for
            # Python
            node = _ast_parse(expr)
            return argname, node, ''  # We consumed the whole expression
        except SyntaxError as error:
            # This probably will be a:
            #    ..., varname: expr...
            #                ^
            offset = error.offset
            while offset > 0 and expr[offset] != ',':
                offset -= 1
            if offset > 0 and expr[offset] == ',':
                return (argname, _ast_parse(expr[:offset]),
                        expr[offset + 1:].lstrip())
            else:
                raise

    class l(object):
        '''A locals implementation that skip some levels up in order to
        protect annotation's own locals.

        '''
        def __init__(self, init={}, skip_levels=5):
            import sys
            # XXX: This code is very fragile, but is the "right" thing to do
            #      in order not to leak implementation-related local variables.
            #      Any lower number will yield wrong results. For instance if
            #      skip_levels is 2, in the following case::
            #
            #          >>> args = 'args'
            #          >>> @annotate('(a: args)')
            #          ... def d():
            #          ...   pass
            #
            #      The annotation for `a` would actually get the tuple
            #      containing the string signature cause in its own
            #      implementation `annotate` uses an `args` local variable::
            #
            #          >>> d.__annotations__
            #          {'a': ('(a: args)',)}
            #
            # XXX: In fact, I should check that this does not create memory
            # references cycles with frames and stuff as noticed in the
            # CPython documentation; notwithstanding that, python's garbage
            # collector may get rid of unreachable objects, even with loops.
            self.f = f = sys._getframe(skip_levels)
            self.f_globals = f.f_globals
            self.d = dict(init)

        def __getitem__(self, key):
            from xoutil.symbols import Unset
            from xoutil.iterators import dict_update_new
            from xoutil.eight import python_version
            d = self.d
            res = d.get(key, Unset)
            f = self.f
            if res is Unset and f:
                f_globals = self.f_globals
                if python_version == 3:
                    # FIXME: This modifies f_globals! Use f_builtins of the
                    # frame.

                    # In Py3k (at least Python 3.2) builtins are not directly
                    # in f_globals but inside a __builtins__ key.
                    builtins = f_globals.get('__builtins__', {})
                    dict_update_new(f_globals, builtins)
                while f and res is Unset:
                    dict_update_new(d, f.f_locals)
                    res = d.get(key, Unset)
                    f = self.f = f.f_back
                if res is Unset and f_globals:
                    dict_update_new(d, f_globals)
                    res = d.get(key, Unset)
                    # At this point there's no use to keep the reference to
                    # frames since we have reached back to the global context,
                    # so it's best to clear of reference to the last frame
                    # in order to keep this CPython-friendly.
                    self.f = None
                    self.f_globals = None
            if res:
                return res
            else:
                raise KeyError

    args, return_annotation = _split_signature(signature)
    while args:
        arg, expr, args = _split_annotation_expression(args)
        code = compile(expr, '', 'eval')
        # Don't put our globals but, just calling-frames globals
        yield arg, eval(code, None, l())
    if return_annotation:
        yield 'return', eval(return_annotation, globals(), l())


@decorator
def annotate(func, signature=None, **keyword_annotations):
    '''Annotates a function with a Python 3k forward-compatible
    ``__annotations__`` mapping.

    See `3107`:pep: for more details about annotations.


    :param signature: A string with the annotated signature of the
        decorated function.

        This string should follow the annotations syntax in `3107`:pep:. But
        there are several deviations from the PEP text:

       - There's no support for the full syntax of Python 2 expressions; in
         particular nested arguments are not supported since they are
         deprecated and are not valid in Py3k.

       - Specifying defaults is no supported (nor needed).  Defaults are
         placed in the signature of the function.

       - In the string it makes no sense to put an argument without an
         annotation, so this will raise an exception (SyntaxError).

    :param keyword_annotations: These are each mapped to a single annotation.

        Since you can't include the 'return' keyword argument for the
        annotation related with the return of the function, we provide several
        alternatives: if any of the following keywords arguments is provided
        (tested in the given order): 'return_annotation', '_return',
        '__return'; then it will be considered the 'return' annotation, the
        rest will be regarded as other annotations.

    In any of the previous cases, you may provide more (or less) annotations
    than possible by following the PEP syntax. This is not considered an error,
    since the PEP allows annotations to be modified by others means.

    If you provide a signature string **and** keywords annotations, the
    keywords will take precedence over the signature::

          >>> @annotate('() -> list', return_annotation=tuple)
          ... def otherfunction():
          ...    pass

          >>> otherfunction.__annotations__.get('return') is tuple
          True

    When parsing the `signature` the locals and globals in the context of the
    declaration are taken into account::

          >>> interface = object # let's mock of ourselves
          >>> class ISomething(interface):
          ...    pass

          >>> @annotate('(a: ISomething) -> ISomething')
          ... def somewhat(a):
          ...     return a

          >>> somewhat.__annotations__.get('a')     # doctest: +ELLIPSIS
          <class '...ISomething'>

    '''
    from xoutil.objects import pop_first_of
    func.__annotations__ = annotations = getattr(func, '__annotations__', {})
    if signature:
        annotations.update({argname: value
                            for argname, value in _parse_signature(signature)})
    probes = ('return_annotation', '_return', '__return')
    return_annotation_kwarg = pop_first_of(keyword_annotations, *probes)
    if return_annotation_kwarg:
        annotations['return'] = return_annotation_kwarg
    annotations.update(keyword_annotations)
    return func



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.names
# ---------------------------------------------------------------------
# Copyright (c) 2013-2017 Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created 2013-04-15

'''A protocol to obtain or manage object names.'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

# FIX: These imports must be local
from xoutil.symbols import Undefined as _undef
from xoutil.eight import base_string


# TODO: This module must be reviewed and deprecate most of it.

def _get_mappings(source):
    '''Return a sequence of mappings from `source`.

    Source could be a stack frame, a single dictionary, or any sequence of
    dictionaries.

    '''
    from collections import Mapping
    if isinstance(source, Mapping):
        return (source,)
    else:
        from xoutil.future.inspect import get_attr_value
        l = get_attr_value(source,  'f_locals', _undef)
        g = get_attr_value(source,  'f_globals', _undef)
        if isinstance(l, Mapping) and isinstance(g, Mapping):
            return (l,) if l is g else (l, g)
        else:
            return tuple(source)


def _key_for_value(source, value, strict=True):
    '''Returns the tuple (key, mapping) where the "value" is found.

    if strict is True, then look first for the same object::
        >>> x = [1]
        >>> y = [1]  #  equal to `x` but not the same
        >>> src = {'x': x, 'y': y}
        >>> search = lambda o, strict=True: _key_for_value(src, o, strict)
        >>> search(x) == search(y)
        False
        >>> search(x, strict=False) == search(y, strict=False)
        True

    This is mainly intended to find object names in stack frame variables.

    Source could be a stack frame, a single dictionary, or any sequence of
    dictionaries.

    '''
    source = _get_mappings(source)
    found, equal = _undef, (None, {})
    i, mapping_count = 0, len(source)
    while found is _undef and (i < mapping_count):
        mapping = source[i]
        keys = list(mapping)
        j, key_count = 0, len(keys)
        while found is _undef and (j < key_count):
            key = keys[j]
            item = mapping[key]
            if item is value:
                found = key, mapping
            elif item == value:
                if strict:
                    equal = key, mapping
                else:
                    found = key, mapping
            j += 1
        i += 1
    return found if found is not _undef else equal


def _get_value(source, key, default=None):
    '''Returns the value for the given `key` in `source` mappings.

    This is mainly intended to obtain object values in stack frame variables.

    Source could be a stack frame, a single dictionary, or any sequence of
    dictionaries.

    '''
    source = _get_mappings(source)
    res = _undef
    i, mapping_count = 0, len(source)
    while res is _undef and (i < mapping_count):
        mapping = source[i]
        res = mapping.get(key, _undef)
        i += 1
    return res if res is not _undef else default


def _get_best_name(names, safe=False, full=False):
    '''Get the best name in the give list of `names`.

    Best names are chosen in the following order (from worst to best):

    - Any string
    - A valid slug
    - A valid protected identifier
    - A valid public identifier
    - A valid full identifier

    If a string in the list `names` contains the substring "%(next)s", then
    the algorithm recurses to find the best name of the remaining names first
    and substitutes the substring with the result, the remaining names are
    then pruned from the search.

    If `safe` is True, returned name must be a valid identifier.  If `full` is
    True (halted if `safe` is not True) then the returned name must a valid
    full identifier.

    '''
    from xoutil.validators import (is_valid_full_identifier,
                                   is_valid_public_identifier,
                                   is_valid_identifier,
                                   is_valid_slug)
    names = list(names)

    def inner(start=0):
        ok, best_idx, best_qlty = start, -1, 0
        i, count = start, len(names)
        assert start < count, 'start is "%s", max is "%s".' % (start, count)
        while i < count:
            name = names[i]
            if '%(next)s' in name:
                next = inner(i + 1)
                names[i] = name % {'next': next}
                count = i + 1
            else:
                if is_valid_slug(name):
                    qlty = 25
                if is_valid_identifier(name):
                    qlty = 75 if is_valid_public_identifier(name) else 50
                elif is_valid_full_identifier(name):
                    qlty = 100
                else:
                    qlty = -25
                if best_qlty <= qlty:
                    best_idx = i
                    best_qlty = qlty
                ok = i
                i += 1
        idx = best_idx if best_idx >= 0 else ok
        return names[idx]
    res = inner()
    if safe:
        # TODO: Improve these methods to return False of reserved identifiers
        is_valid = is_valid_full_identifier if full else is_valid_identifier
        if not is_valid(res):
            from xoutil.string import slugify
            _mark = 'dot_dot_dot'
            full = full and '.' in res
            if full:
                res = res.replace('.', _mark)
            res = slugify(res, '_')
            if full:
                res = res.replace(_mark, '.')
            if not is_valid(res):
                res = '_' + res
    return str(res)


def module_name(item):
    '''Returns the full module name where the given object is declared.

    Examples::

       >>> module_name(module_name)
       'xoutil.names'

       >>> from xoutil.symbols import Unset
       >>> module_name(Unset)
       'xoutil.symbols'

    '''
    from xoutil.future.inspect import get_attr_value
    if item is None:
        res = ''
    elif isinstance(item, base_string):
        res = item
    else:
        res = get_attr_value(item, '__module__', None)
        if res is None:
            res = get_attr_value(type(item), '__module__', '')
    if res.startswith('__') or res in ('builtins', 'exceptions', '<module>'):
        res = ''
    return str(res)


def simple_name(item, join=True):
    '''Returns the simple name for the given object.

    :param join: If False, only the object inner name is returned; if it is a
           callable is used similar to a string join receiving a tuple of
           (module-name, inner-name) as argument; True means (is equivalent
           to)::

             join = lambda arg: '.'.join(arg).strip('.')

           For example, use ``lambda arg: arg`` to return the 2-tuple itself.

           See `module_name`:func: for more information when a not False value
           is used.

    Examples::

       >>> simple_name(simple_name)
       'xoutil.names.simple_name'

       >>> from xoutil.symbols import Unset
       >>> simple_name(Unset)
       'xoutil.symbols.Unset'

    This function is intended for named objects (those with the `__name__`
    attribute), if an object without standard name is used, the type name is
    returned instead; for example::

        >>> simple_name(0)
        'int'

    To get a name in a more precise way, use `nameof`:func:.

    '''
    # TODO: deprecate `join` argument
    from xoutil.future.inspect import safe_name
    singletons = (None, True, False, Ellipsis, NotImplemented)
    res = next((str(s) for s in singletons if s is item), None)
    if res is None:
        res = safe_name(item)
        if res is None:
            item = type(item)
            res = safe_name(item)
        if join:
            if join is True:
                def join(arg):
                    return str('.'.join(arg).strip('.'))
            res = join((module_name(item), res))
    return res


def nameof(*args, **kwargs):
    '''Obtain the name of each one of a set of objects.

    .. versionadded:: 1.4.0

    .. versionchanged:: 1.6.0

       - Keyword arguments are now keyword-only arguments.

       - Support for several objects

       - Improved the semantics of parameter `full`.

       - Added the `safe` keyword argument.

    If no object is given, None is returned; if only one object is given, a
    single string is returned; otherwise a list of strings is returned.

    The name of an object is normally the variable name in the calling stack.

    If the object is not present calling frame, up to five frame levels are
    searched.  Use the `depth` keyword argument to specify a different
    starting point and the search will proceed five levels from this frame up.

    If the same object has several good names a single one is arbitrarily
    chosen.

    Good names candidates are retrieved based on the keywords arguments
    `full`, `inner`, `safe` and `typed`.

    If `typed` is True and the object is not a type name or a callable (see
    `xoutil.future.inspect.safe_name`:func:), then the `type` of the object is
    used instead.

    If `inner` is True we try to extract the name by introspection instead of
    looking for the object in the frame stack.

    If `full` is True the full identifier of the object is preferred.  In this
    case if `inner` is False the local-name for the object is found.  If
    `inner` is True, find the import-name.

    If `safe` is True, returned value is converted -if it is not- into a valid
    Python identifier, though you should not trust this identifier resolves to
    the value.

    See `the examples in the documentation <name-of-narrative>`:ref:.

    '''
    # XXX: The examples are stripped from here.  Go the documentation page.
    from numbers import Number
    from xoutil.eight import range
    from xoutil.future.inspect import safe_name
    arg_count = len(args)
    names = [[] for i in range(arg_count)]

    class vars:
        '`nonlocal` simulation'
        params = kwargs
        idx = 0

    def grant(name=None, **again):
        if name:
            names[vars.idx].append(name)
            assert len(names[vars.idx]) < 5
        if again:
            vars.params = dict(kwargs, **again)
        else:
            vars.params = kwargs
            vars.idx += 1

    def param(name, default=False):
        return vars.params.get(name, default)

    while vars.idx < arg_count:
        item = args[vars.idx]
        if param('typed') and not safe_name(item):
            item = type(item)
        if param('inner'):
            res = safe_name(item)
            if res:
                if param('full'):
                    head = module_name(item)
                    if head:
                        res = '.'.join((head, res))
                grant(res)
            elif isinstance(item, (base_string, Number)):
                grant(str(item))
            else:
                grant('@'.join(('%(next)s', hex(id(item)))), typed=True)
        else:
            import sys
            sf = sys._getframe(param('depth', 1))
            try:
                i, LIMIT, res = 0, 5, _undef
                _full = param('full')
                while not res and sf and (i < LIMIT):
                    key, mapping = _key_for_value(sf, item)
                    if key and _full:
                        head = module_name(_get_value(mapping, '__name__'))
                        if not head:
                            head = module_name(sf.f_code.co_name)
                        if not head:
                            head = module_name(item) or None
                    else:
                        head = None
                    if key:
                        res = key
                    else:
                        sf = sf.f_back
                        i += 1
            finally:
                # TODO: on "del sf" Python says "SyntaxError: can not delete
                # variable 'sf' referenced in nested scope".
                sf = None
            if res:
                grant('.'.join((head, res)) if head else res)
            else:
                res = safe_name(item)
                if res:
                    grant(res)
                else:
                    grant(None, inner=True)
    for i in range(arg_count):
        names[i] = _get_best_name(names[i], safe=param('safe'))
    if arg_count == 0:
        return None
    elif arg_count == 1:
        return names[0]
    else:
        return names


def identifier_from(*args):
    '''Build an valid identifier from the name extracted from an object.

    .. versionadded:: 1.5.6

    First, check if argument is a type and then returns the name of the type
    prefixed with `_` if valid; otherwise calls `nameof` function repeatedly
    until a valid identifier is found using the following order logic:
    ``inner=True``, without arguments looking-up a variable in the calling
    stack, and ``typed=True``.  Returns None if no valid value is found.

    Examples::

        >>> identifier_from({})
        'dict'

    '''
    if len(args) == 1:
        from xoutil.validators.identifiers import is_valid_identifier as valid
        from xoutil.future.inspect import get_attr_value
        res = None
        if isinstance(args[0], type):
            aux = get_attr_value(args[0], '__name__', None)
            if valid(aux):
                res = str('_%s' % aux)
        if res is None:
            tests = ({'inner': True}, {}, {'typed': True})
            names = (nameof(args[0], depth=2, **test) for test in tests)
            res = next((name for name in names if valid(name)), None)
        return res
    else:
        msg = 'identifier_from() takes exactly 1 argument (%s given)'
        raise TypeError(msg % len(args))


class namelist(list):
    '''Similar to list, but only intended for storing object names.

    Constructors:

    * namelist() -> new empty list
    * namelist(collection) -> new list initialized from collection's items
    * namelist(item, ...) -> new list initialized from severals items

    Instances can be used as decorators to store names of module items
    (functions or classes)::

        >>> __all__ = namelist()
        >>> @__all__
        ... def foobar(*args, **kwargs):
        ...     'Automatically added to this module "__all__" names.'

        >>> 'foobar' in __all__
        True

    '''
    def __init__(self, *args):
        if len(args) == 1:
            from types import GeneratorType as gtype
            if isinstance(args[0], (tuple, list, set, frozenset, gtype)):
                args = args[0]
        super(namelist, self).__init__(nameof(arg, depth=2) for arg in args)

    def __add__(self, other):
        other = [nameof(item, depth=2) for item in other]
        return super(namelist, self).__add__(other)

    __iadd__ = __add__

    def __contains__(self, item):
        return super(namelist, self).__contains__(nameof(item, inner=True))

    def append(self, value):
        '''l.append(value) -- append a name object to end'''
        super(namelist, self).append(nameof(value, depth=2))
        return value    # What allow to use its instances as a decorator

    __call__ = append

    def extend(self, items):
        '''l.extend(items) -- extend list by appending items from the iterable
        '''
        items = (nameof(item, depth=2) for item in items)
        return super(namelist, self).extend(items)

    def index(self, value, *args):
        '''l.index(value, [start, [stop]]) -> int -- return first index of name

        Raises ValueError if the name is not present.

        '''
        return super(namelist, self).index(nameof(value, depth=2), *args)

    def insert(self, index, value):
        '''l.insert(index, value) -- insert object before index
        '''
        return super(namelist, self).insert(index, nameof(value, depth=2))

    def remove(self, value):
        '''l.remove(value) -- remove first occurrence of value

        Raises ValueError if the value is not present.

        '''
        return list.remove(self, nameof(value, depth=2))


class strlist(list):
    '''Similar to list, but only intended for storing ``str`` instances.

    Constructors:
      * strlist() -> new empty list
      * strlist(collection) -> new list initialized from collection's items
      * strlist(item, ...) -> new list initialized from severals items

    Last versions of Python 2.x has a feature to use unicode as standard
    strings, but some object names can be only ``str``. To be compatible with
    Python 3.x in an easy way, use this list.

    '''
    def __init__(self, *args):
        if len(args) == 1:
            from types import GeneratorType as gtype
            if isinstance(args[0], (tuple, list, set, frozenset, gtype)):
                args = args[0]
        super(strlist, self).__init__(str(arg) for arg in args)

    def __add__(self, other):
        other = [str(item) for item in other]
        return super(strlist, self).__add__(other)

    __iadd__ = __add__

    def __contains__(self, item):
        return super(strlist, self).__contains__(str(item))

    def append(self, value):
        '''l.append(value) -- append a name object to end'''
        super(strlist, self).append(str(value))
        return value    # What allow to use its instances as a decorator

    __call__ = append

    def extend(self, items):
        '''l.extend(items) -- extend list by appending items from the iterable
        '''
        items = (str(item) for item in items)
        return super(strlist, self).extend(items)

    def index(self, value, *args):
        '''l.index(value, [start, [stop]]) -> int -- return first index of name

        Raises ValueError if the name is not present.

        '''
        return super(strlist, self).index(str(value), *args)

    def insert(self, index, value):
        '''l.insert(index, value) -- insert object before index
        '''
        return super(strlist, self).insert(index, str(value))

    def remove(self, value):
        '''l.remove(value) -- remove first occurrence of value

        Raises ValueError if the value is not present.

        '''
        return list.remove(self, str(value))


# Theses tests need to be defined in this module to test relative imports.
# Otherwise the `tests/` directory would need to be a proper package.

import unittest as _utest
from xoutil.symbols import Unset as _Unset   # Use a tier 0 module!


class TestRelativeImports(_utest.TestCase):
    RelativeUnset = _Unset
    AbsoluteUndefined = _undef

    def test_relative_imports(self):
        self.assertEquals(nameof(self.RelativeUnset), '_Unset')
        self.assertEquals(nameof(self.RelativeUnset, inner=True), 'Unset')
        self.assertEquals(nameof(self.RelativeUnset, full=True),
                          'xoutil.names._Unset')  # Even relative imports are
                                                  # resolved properly with
                                                  # `full=True`
        self.assertEquals(nameof(self.AbsoluteUndefined, full=True),
                          'xoutil.names._undef')

# Don't delete the _Unset name, so that the nameof inside the test could find
# them in the module.
del _utest



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.tasking
# ---------------------------------------------------------------------
# Copyright (c) 2015-2017 Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2015-10-25

'''Multitasking and concurrent programming tools.

.. warning:: Experimental.  API is not settled.

.. versionadded:: 1.8.0

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

import sys

# TODO: Must be implemented using `xoutil.api` mechanisms for correct driver
# determination, in this case "thread-local data".
if 'greenlet' in sys.modules:
    from ._greenlet_local import local    # noqa
else:
    try:
        from threading import local    # noqa
    except ImportError:
        from dummy_threading import local    # noqa

del sys


class AutoLocal(local):
    '''Initialize thread-safe local data in one shoot.

    Typical use is::

        >>> from xoutil.tasking import AutoLocal
        >>> context = AutoLocal(cause=None, traceback=None)

    When at least one attribute is given, ``del AutoLocal`` it's executed
    automatically avoiding this common statement at the end of your module.

    '''
    def __init__(self, **attrs):
        import sys
        super(AutoLocal, self).__init__()
        for attr in attrs:
            setattr(self, attr, attrs[attr])
        if attrs:
            g = sys._getframe(1).f_globals
            tname = type(self).__name__
            if tname in g:
                del g[tname]



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.future
# ---------------------------------------------------------------------
# Copyright (c) 2016-2017 Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2016-09-10

'''Extend standard modules including "future" features in current versions.

Version 3 introduce several concepts in standard modules.  Sometimes these
features are implemented in the evolution of 2.7.x versions.  By using
sub-modules, these differences can be avoided transparently.  For example, you
can import `xoutil.future.collections.UserDict`:class: in any version, that
it's equivalent to Python 3 `collections.UserDict`:class:, but it don't exists
in Python 2.

.. versionadded:: 1.7.2

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

# Ensure that, inside all 'xoutil.future' sub-modules, the ABC register method
# can be used as a decorator in all Python versions.
from xoutil.eight import abc
del abc



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.fp.option
# ---------------------------------------------------------------------
# Copyright (c) 2015-2017 Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2015-12-07

'''Functional Programming *Option Type* definition.

In Programming, and Type Theory, an *option type*, or *maybe type*, represents
encapsulation of an optional value; e.g., it is used in functions which may or
may not return a meaningful value when they are applied.

It consists of either a constructor encapsulating the original value ``x``
(written ``Just x`` or ``Some x``) or an empty constructor (called *None* or
*Nothing*).  Outside of functional programming, these are known as *nullable
types*.

In our case *option type* will be the `Maybe`:class: class (the equivalent of
`Option` in *Scala Programming Language*), the wrapper for valid values will
be the `Just`:class: class (equivalent of `Some` in *Scala*); and the wrapper
for invalid values will be the `Wrong`:class: class.

Instead of *None* or *Nothing*, `Wrong` is used because two reasons:
(1) already existence of `None` special Python value, and (2) `Wrong`:class:
also wraps incorrect values and can have several instances (not only a *null*
value).

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)


class Maybe(object):
    '''Wrapper for optional values.

    The Maybe type encapsulates an optional value.  A value of type
    ``Maybe a`` either contains a value of type ``a`` (represented as
    ``Just a``), or it is empty (represented as ``Nothing``).  Using `Maybe``
    is a good way to deal with errors or exceptional cases without resorting
    to drastic measures such as error. In this implementation we make a
    variation where a ``Wrong`` object represents a missing (with special
    value ``Nothing``) or an improper value (including errors).

    See descendant classes `Just`:class: and `Wrong`:class: for more
    information.

    This implementation combines ``Maybe`` and ``Either`` Haskell data types.
    ``Maybe`` is a means of being explicit that you are not sure that a
    function will be successful when it is executed.  Conventionally, the
    usage of ``Either`` for errors uses ``Right`` when the computation is
    successful, and ``Left`` for failing scenarios.

    In this implementation, `Just`:class` us used for equivalence with both
    Haskell ``Just`` and ``Right`` types; `Wrong`:class: is used with the
    special value ``Nothing`` and to encapsulate errors or incorrect values
    (Haskell ``Left``).

    Haskell::

      data Maybe a = Nothing | Just a

      either :: (a -> c) -> (b -> c) -> Either a b -> c

    Case analysis for the Either type.  If the value is Left a, apply the
    first function to a; if it is Right b, apply the second function to b.

    '''
    __slots__ = 'inner'
    _singletons = [None, None, None]    # False, True, None

    def __new__(cls, *args):
        default = cls is Just
        if len(args) == 0:
            arg = default
        elif len(args) == 1:
            arg = args[0]
        else:
            msg = '{}: receive too many arguments "{}"'
            raise TypeError(msg.format(cls.__name__, len(args)))
        if arg is default or arg is None and cls is Wrong:
            idx = 2 if arg is None else arg
            if cls._singletons[idx] is None:
                self = super(Maybe, cls).__new__(cls)
                self.inner = arg
                cls._singletons[idx] = self
            return cls._singletons[idx]
        elif cls is Maybe:
            return (Just if arg else Wrong)(arg)
        elif isinstance(arg, cls):
            return arg
        elif not isinstance(arg, Maybe):
            self = super(Maybe, cls).__new__(cls)
            self.inner = arg
            return self
        else:
            msg = 're-wrapping inverted value: {}({})'
            raise ValueError(msg.format(cls.__name__, arg))

    def __init__(self, *args):
        pass

    def __nonzero__(self):
        return isinstance(self, Just)
    __bool__ = __nonzero__

    def __str__(self):
        return '{}({!r})'.format(type(self).__name__, self.inner)
    __repr__ = __str__

    def __eq__(self, other):
        return (isinstance(other, type(self)) and self.inner == other.inner or
                self.inner is other)    # TODO: check if `==` instead `is`

    def __ne__(self, other):
        return not (self == other)

    @classmethod
    def triumph(cls, value):
        '''Coerce to a logical Boolean value.

        A wrapper `Just`:class: is logically true, and `Wrong` is false.

        For example::

            >>> Just.triumph([1])
            [1]

            >>> Just.triumph([])
            Just([])

            >>> Wrong.triumph([1])
            Wrong([1])

            >>> Wrong.triumph([])
            []

        '''
        default = cls is Just
        if bool(value) is default:
            idx = 2 if arg is None else arg
            if cls._singletons[idx] is None:
                self = super(Maybe, cls).__new__(cls)
                self.inner = arg
                cls._singletons[idx] = self
            return cls._singletons[idx]
        elif cls is Maybe:
            return (Just if arg else Wrong)(arg)
        elif isinstance(arg, cls):
            return arg
        elif not isinstance(arg, Maybe):
            self = super(Maybe, cls).__new__(cls)
            self.inner = arg
            return self
        else:
            msg = 're-wrapping inverted value: {}({})'
            raise ValueError(msg.format(cls.__name__, arg))

    @classmethod
    def compel(cls, value):
        '''Coerce to the correspondent logical Boolean value.

        `Just`:class: is logically true, and `Wrong` is false.

        For example::

            >>> Just.compel([1])
            [1]

            >>> Just.compel([])
            Just([])

            >>> Wrong.compel([1])
            Wrong([1])

            >>> Wrong.compel([])
            []

        '''
        from xoutil.eight import type_name
        if cls is not Maybe:
            test = cls is Just
            dual = Wrong if test else Just
            if bool(value) is test:
                return value
            elif not isinstance(value, dual):
                return cls(value)
            else:
                msg = '''a "{}" value can't be coerced to "{}"'''
                raise TypeError(msg.format(type_name(value), cls.__name__))
        else:
            raise TypeError('''don't call at Maybe base level''')

    @classmethod
    def choose(cls, *types):
        '''Decorator to force `Maybe` values constraining to expecting types.

        For example, a function that return a collection (tuple or list) if
        valid or False if not, if not decorated could be ambiguous for an
        empty collection::

            >>> @Just.choose(tuple, list)
            ... def check_range(values, min, max):
            ...     if isinstance(values, (tuple, list)):
            ...         return [v for v in values if min <= v <= max]
            ...     else:
            ...         return False

            >>> check_range(range(10), 7, 17)
            [7, 8, 9]

            >>> check_range(range(10), 17, 27)
            Just([])

            >>> check_range(set(range(10)), 7, 17)
            False

        '''
        pass


class Just(Maybe):
    '''A wrapper for valid results.'''
    __slots__ = ()


class Wrong(Maybe):
    '''A wrapper for invalid results.

    When encapsulation errors, the current trace-back is properly encapsulated
    using `xoutil.eight.exceptions`:mod: module features.

    # TODO: Use `naught` or `Left` instead.

    '''
    __slots__ = ()

    def __new__(cls, *args):
        self = super(Wrong, cls).__new__(cls, *args)
        if isinstance(self.inner, BaseException):
            from xoutil.eight.exceptions import catch
            self.inner = catch(self.inner)
        return self


def take(value):
    '''Extract a value.'''
    return value.inner if isinstance(value, Maybe) else value


# ---- special singletons ----

#: A `Wrong`:class: special singleton encapsulating the `False` value.
false = Wrong()

#: A `Just`:class: special singleton encapsulating the `True` value.
true = Just()

#: A `Wrong`:class: special singleton encapsulating the `None` value.
none = Wrong(None)



# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.bound
# ---------------------------------------------------------------------
# Copyright (c) 2014-2017 Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2014-07-21

'''Helpers for bounded execution of co-routines.

Example::

    >>> def fibonacci():
    ...   a, b = 1, 1
    ...   while True:
    ...       yield a
    ...       a, b = b, a + b

This function yields forever.  This module allows to get instances of that
function that run until a boundary condition is met.  For instance, the
`times`:func: boundary stops after a given numbers of results are generated::

    >>> fib8 = times(8)(fibonacci)
    >>> fib8()   # the 8th fibonacci number is
    21

This is repeatable::

    >>> fib8()   # the 8th fibonacci number is
    21

    >>> fib8()   # the 8th fibonacci number is
    21

Unless you pass in a generator::

    >>> fib8 = times(8)(fibonacci())
    >>> fib8()
    21

    >>> fib8() is None
    True

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from types import GeneratorType
from xoutil.decorator.meta import decorator
from xoutil.eight.meta import metaclass


class BoundedType(type):
    '''A bounded generator/function.'''
    pass


class Bounded(metaclass(BoundedType)):
    '''The bounded function.

    This is the result of applying a `boundary definition` to an `unbounded
    function` (or generator).

    If `target` is a function this instance can be called several times.  If
    it's a generator then it will be closed after either calling
    (``__call__``) this instance, or consuming the generator given by
    `generate`:meth:.

    '''
    def __init__(self, target):
        self.target = target

    # The following two methods are actually implemented as closures in the
    # apply method of BoundaryCondition.  Nevertheless, they are documented
    # here as an API promise.
    def __call__(self, *args, **kwargs):
        '''Return the last value from the underlying `bounded generator`.

        '''
        raise NotImplementedError()

    def generate(self, *args, **kwargs):
        '''Return the `bounded generator`.

        This method exposes the `bounded generator`.  This allows you to "see"
        all the values yielded by the `unbounded generator` up to the point
        when the boundary condition is met.

        '''
        raise NotImplementedError()


class BoundaryCondition(object):
    '''Embodies the boundary protocol.

    The `definition` argument must a function that implements a `boundary
    definition`.  This function may take arguments to initialize the state of
    the boundary condition.

    Instances are callables that will return a `Bounded`:class: subclass
    specialized with the application of the `boundary condition` to a given
    unbounded function (`target`).  For instance, ``times(6)`` returns a
    class, that when instantiated with a `target` represents the bounded
    function that takes the 6th valued yielded by target.

    If the `definition` takes no arguments for initialization you may pass the
    `target` directly.  This is means that if `__call__`:func: receives
    arguments they will be used to instantiate the `Bounded`:class: subclass,
    ie. this case allows only a single argument `target`.

    If `errors` is not None it should be a tuple of exceptions to catch and
    throw inside the boundary condition definition.  Other exceptions, beside
    GeneratorExit and StopIteration, are not handled (so the bubble up).  See
    `until_error`:func:.

    '''
    def __new__(cls, definition, name=None, errors=None):
        from types import FunctionType
        if not isinstance(definition, FunctionType):
            raise TypeError('"definition" must be a function')
        if not name:
            from xoutil.names import nameof
            name = nameof(definition, inner=True, full=True)
        result = super(BoundaryCondition, cls).__new__(cls)
        result.name = name  # needs to be set here or it'll be None
        return result

    def __init__(self, definition, name=None, errors=None):
        from inspect import getargspec
        spec = getargspec(definition)
        self.args = spec[0]
        self.defaults = spec[3]
        self.varargs = spec[1]
        self.varkwargs = spec[2]
        self.definition = definition
        if not errors:
            errors = (Exception, )
        self.errors = errors

    def __str__(self):
        return str('boundary %s(...)' % self.name)

    def __repr__(self):
        return str(self)

    @property
    def receive_args(self):
        return self.args or self.defaults or self.varargs or self.varkwargs

    def apply(self, args, kwargs):
        def execute(boundary, unbounded, initial):
            '''Executes the unbounded generator guarded by a boundary condition.

            `boundary` is the boundary condition. `unbounded` is the unbounded
            generator.  Both must be generators.

            `initial` is the tuple of ``(args, kwargs)`` passed when calling
            the unbounded function or None.

            This function is used in the (closure) `generate` method of the
            Bounded subclass returned by `apply`.  It contains the core
            algorithm that interleaves the boundary condition with the
            unbounded generator.

            '''
            try:
                next(boundary)  # Initialize the boundary condition
                stop = boundary.send(initial)
            except StopIteration:
                raise RuntimeError(
                    'Invalid boundary definition "%r"' % self.definition
                )
            try:
                while stop is not True:
                    try:
                        data = next(unbounded)
                        yield data
                    except (GeneratorExit, StopIteration):
                        stop = True
                    except self.errors as error:
                        stop = boundary.throw(error)
                    else:
                        try:
                            stop = boundary.send(data)
                        except StopIteration:
                            raise RuntimeError(
                                'Invalid boundary definition "%r"' %
                                self.definition
                            )
            finally:
                boundary.close()
                unbounded.close()

        class bounded(Bounded):
            @classmethod
            def build_pred(boundedcls):
                return self.build_generator(args, kwargs)

            def generate(me, *args, **kwargs):
                target = me.target
                if isinstance(target, GeneratorType):
                    return execute(me.build_pred(), target, None)
                else:
                    generator = target(*args, **kwargs)
                    return execute(me.build_pred(), generator, (args, kwargs))

            def __call__(me, *args, **kwargs):
                data = None
                for data in me.generate(*args, **kwargs):
                    pass
                return data

        return bounded  # return from apply()

    def build_generator(self, args, kwargs):
        if self.receive_args:
            generator = self.definition(*args, **kwargs)
        else:
            generator = self.definition()
        return generator

    def __call__(self, *args, **kwargs):
        if self.receive_args:
            return self.apply(args, kwargs)
        elif args or kwargs:
            result = self.apply((), {})(*args, **kwargs)
            if len(args) == 1:
                from functools import update_wrapper
                update_wrapper(result, args[0])
            return result
        else:
            return self.apply((), {})


@decorator
def boundary(definition, name=None, base=BoundaryCondition, errors=None):
    '''Helper to define a boundary condition.

    The `definition` must be a function that returns a generator.  The
    following rules **must be** followed.  Collectively these rules are called
    the `boundary protocol`.

    - The `boundary definition` will yield True when and only when the
      boundary condition is met.  Only the value True will signal the boundary
      condition.

    - The `boundary definition` must yield at least 2 times:

      - First it will be called its ``next()`` method to allow for
        initialization of internal state.

      - Immediately after, it will be called its ``send()`` passing the tuple
        ``(args, kwargs)`` with the arguments passed to the `unbounded
        function`.  At this point the boundary definition may yield True to
        halt the execution.  In this case, the `unbounded generator` won't be
        asked for any value.

    - The `boundary definition` must yield True before terminating with a
      StopIteration.  For instance the following definition is invalid cause
      it ends without yielding True::

          @boundary
          def invalid():
              yield
              yield False

    - The `boundary definition` must deal with GeneratorExit exceptions
      properly since we call the ``close()`` method of the generator upon
      termination.  Termination occurs when the `unbounded generator` stops by
      any means, even when the boundary condition yielded True or the
      generator itself is exhausted or there's an error in the generator.

      Both `whenall`:func: and `whenany`:func: call the ``close()`` method of
      all their subordinate boundary conditions.

      Most of the time this reduces to *not* catching GeneratorExit
      exceptions.

    A RuntimeError may happen if any of these rules is not followed by the
    `definition`.  Furthermore, this error will occur when invoking the
    `bounded function` and not when applying the boundary to the `unbounded
    generator`.

    '''
    from functools import update_wrapper
    result = base(definition, name=name, errors=errors)
    return update_wrapper(result, definition)


@boundary
def timed(maxtime):
    '''Becomes True after a given amount of time.

    The bounded generator will be allowed to yields values until the `maxtime`
    time frame has elapsed.

    Usage::

         @timed(timedelta(seconds=60))
         def do_something_in_about_60s():
             while True:
                 yield

    .. note:: This is a very soft limit.

       We can't actually guarrant any enforcement of the time limit.  If the
       bounded generator takes too much time or never yields this predicated
       can't do much.  This usually helps with batch processing that must not
       exceed (by too much) a given amount of time.

    The timer starts just after the ``next()`` function has been called for
    the predicate initialization.  So if the `maxtime` given is too short this
    predicated might halt the execution of the bounded function without
    allowing any processing at all.

    If `maxtime` is not a timedelta, the timedelta will be computed as
    ``timedelta(seconds=maxtime)``.

    '''
    from datetime import datetime, timedelta
    if isinstance(maxtime, timedelta):
        bound = maxtime
    else:
        bound = timedelta(seconds=maxtime)
    start = datetime.now()
    yield False  # Deal with next-send calling scheme for boundaries
    while datetime.now() - start < bound:
        yield False
    yield True   # Or we're not compliant with the boundary protocol.


@boundary
def times(n):
    '''Becomes True after a given after the `nth` item have been produced.'''
    passed = 0
    yield False
    while passed < n:
        yield False
        passed += 1
    yield True


@boundary
def accumulated(mass, *attrs, **kwargs):
    '''Becomes True after accumulating a given "mass".

    `mass` is the maximum allowed to accumulate.  This is usually a positive
    number.  Each value produced by the `unbounded generator` is added
    together.  Yield True when this amount to more than the given `mass`.

    If any `attrs` are provided, they will be considered attributes (or keys)
    to search inside the yielded data from the bounded function.  If no
    `attrs` are provided the whole data is accumulated, so it must allow
    addition.  The attribute to be summed is extracted with
    `~xoutil.objects.get_first_of`:func:, so only the first attribute found is
    added.

    If the keyword argument `initial` is provided the accumulator is
    initialized with that value.  By default this is 0.

    '''
    from xoutil.objects import get_first_of
    accum = kwargs.pop('initial', 0)
    if kwargs:
        raise TypeError('Invalid keyword arguments %r' % kwargs.keys())
    yield False
    while accum < mass:
        data = yield False
        accum += get_first_of(data, *attrs, default=data)
    yield True


@boundary
def pred(func, skipargs=True):
    '''Allow "normal" functions to engage within the boundary protocol.

    `func` should take a single argument and return True if the boundary
    condition has been met.

    If `skipargs` is True then function `func` will not be called with the
    tuple ``(args, kwargs)`` upon initialization of the boundary, in that case
    only yielded values from the `unbounded generator` are passed.  If you
    need to get the original arguments, set `skipargs` to False, in this case
    the first time `func` is called will be passed a single argument ``(arg,
    kwargs)``.

    Example::

      >>> @pred(lambda x: x > 10)
      ... def fibonacci():
      ...     a, b = 1, 1
      ...     while True:
      ...        yield a
      ...        a, b = b, a + b

      >>> fibonacci()
      13

    '''
    sentinel = object()
    data = yield False
    if skipargs:
        data = sentinel
    while data is sentinel or not func(data):
        data = yield False
    yield True


def until_errors(*errors, **kwargs):
    '''Becomes True after any of `errors` has been raised.

    Any other exceptions (except GeneratorExit) is propagated.  You must pass
    at least an error.

    Normally this will allow some possibly long jobs to be interrupted
    (SoftTimeLimitException in celery task, for instance) but leave some time
    for the caller to clean up things.

    It's assumed that your job can be properly *finalized* after any of the
    given exceptions has been raised.

    :keyword on_error: A callable that will only be called if the boundary
                       condition is ever met, i.e if any of `errors` was
                       raised.  The callback is called before yielding True.

    .. versionadded:: 1.7.2

    .. versionchanged:: 1.7.5 Added the keyword argument `on_error`.

    '''
    if not errors:
        raise TypeError('catch must be called with at least an exception')
    elif any(not issubclass(e, Exception) for e in errors):
        raise TypeError(
            'catch must be called only with subclasses of Exception'
        )
    if any(issubclass(e, GeneratorExit) for e in errors):
        raise TypeError('You cannot catch GeneratorExit')
    on_error = kwargs.pop('on_error', None)
    if kwargs:
        raise TypeError('Invalid keyword arguments: %s' % ', '.join(kwargs))

    @boundary(errors=errors)
    def _catch():
        yield False
        try:
            while True:
                yield False
        except errors:
            if on_error is not None:
                on_error()
            yield True
    return _catch()


def until(**kwargs):
    '''An idiomatic alias to other boundary definitions.

    - ``until(maxtime=n)`` is the same as ``timed(n)``.

    - ``until(times=n)`` is the same as ``times(n)``.

    - ``until(pred=func, skipargs=skip)`` is the same as
      ``pred(func, skipargs=skip)``.

    - ``until(errors=errors, **kwargs)`` is the same as
      ``until_errors(*errors, **kwargs)``.

    - ``until(accumulate=mass, path=path, initial=initial)`` is the same as
       ``accumulated(mass, *path.split('.'), initial=initial)``

    .. warning:: You cannot mix many calls.

    .. versionadded:: 1.7.2

    '''
    maxtime = kwargs.pop('maxtime', None)
    if maxtime:
        return timed(maxtime, **kwargs)
    n = kwargs.pop('times', None)
    if n:
        return times(n, **kwargs)
    func = kwargs.pop('pred', None)
    if func:
        return pred(func, **kwargs)
    errors = kwargs.pop('errors', None)
    if errors:
        return until_errors(*errors, **kwargs)
    mass = kwargs.pop('accumulate', None)
    if mass:
        path = kwargs.pop('path', None)
        if path:
            return accumulated(mass, *path.split('.'), **kwargs)
        else:
            return accumulated(mass, **kwargs)
    raise TypeError


class HighLevelBoundary(BoundaryCondition):
    '''Boundary class for high-level boundary conditions.

    The `apply` method of this only accepts the `args`, which must be
    BoundaryCondition objects or BoundedType objects (ie. an instance of a
    boundary condition), then it replaces the normal boundary condition for
    that of the high-level given the subordinate definitions.

    '''

    def apply(self, boundaries, kwargs):
        assert boundaries and not kwargs
        base = super(HighLevelBoundary, self).apply(boundaries, kwargs)

        class rebounded(base):
            @classmethod
            def build_pred(cls):
                from types import FunctionType, GeneratorType
                subordinates = []
                for bound in boundaries:
                    if isinstance(bound, FunctionType):
                        bound = boundary(bound)
                    elif isinstance(bound, GeneratorType):
                        gen = bound  # get a copy for the lambda below
                        bound = boundary(lambda: gen)
                    if isinstance(bound, BoundaryCondition):
                        if bound.receive_args:
                            raise TypeError(
                                '"%s" must be initialized' % bound.name
                            )
                        bound = bound.apply((), {})
                    if isinstance(bound, BoundedType):
                        sub = bound.build_pred()
                    else:
                        raise TypeError('Invalid argument "%r"' % bound)
                    subordinates.append(sub)
                return self.definition(*subordinates)
        return rebounded


@boundary(base=HighLevelBoundary)
def whenall(*subordinates):
    '''An AND-like boundary condition.

    It takes several boundaries and returns a single one that behaves like the
    logical AND i.e, will yield True when **all** of its subordinate boundary
    conditions have yielded True.

    It ensures that once a subordinate yields True it won't be sent more data,
    no matter if other subordinates keep on running and consuming data.

    Calls ``close()`` of all subordinates upon termination.

    Each `boundary` should be either:

    - A "bare" boundary definition that takes no arguments.

    - A boundary condition (i.e an instance of `BoundaryCondition`:class:).
      This is result of calling a boundary definition.

    - A generator object that complies with the boundary protocol.  This
      cannot be tested upfront, a misbehaving generator will cause a
      RuntimeError if a boundary protocol rule is not followed.

    Any other type is a TypeError.

    '''
    preds = list(subordinates)  # a copy of the list
    for pred in preds:
        next(pred)
    try:
        while preds:  # out of preds it means all have yielded True
            data = yield False
            i = 0
            while preds and i < len(preds):
                pred = preds[i]
                try:
                    res = pred.send(data)
                except StopIteration:
                    raise RuntimeError('Invalid predicated in %r' % preds)
                else:
                    if res is True:
                        del preds[i]  # no more send() for this pred
                    else:
                        i += 1
        yield True
    except GeneratorExit:
        pass
    for pred in subordinates:
        pred.close()


@boundary(base=HighLevelBoundary)
def whenany(*preds):
    '''An OR-like boundary condition.

    It takes several boundaries and returns a single one that behaves like the
    logical OR, i.e, will yield True when **any** of its subordinate boundary
    conditions yield True.

    Calls ``close()`` of all subordinates upon termination.

    Each `boundary` should be either:

    - A "bare" boundary definition that takes no arguments.

    - A boundary condition (i.e an instance of `BoundaryCondition`:class:).
      This is result of calling a boundary definition.

    - A generator object that complies with the boundary protocol.  This
      cannot be tested upfront, a misbehaving generator will cause a
      RuntimeError if a boundary protocol rule is not followed.

    Any other type is a TypeError.

    '''
    for pred in preds:
        next(pred)
    stop = False
    try:
        while stop is not True:
            data = yield stop
            i, top = 0, len(preds)
            while not stop and i < top:
                pred = preds[i]
                try:
                    stop = stop or pred.send(data)
                except StopIteration:
                    raise RuntimeError('Invalid predicated in %r' % preds)
                else:
                    i += 1
        yield stop
    except GeneratorExit:
        pass
    for pred in preds:
        pred.close()


del decorator, metaclass



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.eight.mixins
# ---------------------------------------------------------------------
# Copyright (c) 2015-2017 Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2015-10-29

'''Two functions to create helper classes and mixins.

This module is in the `eight` context because these two functions depend on
several concepts that are different in Python 2 and 3.

- `helper_class`:func: creates a base class that represent a meta-class.  For
  example (only for Python 3), `xoutil.eight.abc.ABC` is different to
  `abc.ABC`::

    >>> from xoutil.eight.abc import ABC, ABCMeta
    >>> class One(ABC):
    ...     pass
    >>> One.__bases__ == (ABC, )
    False
    >>> One.__bases__ == (Mixin, )
    True

    >>> from abc import ABC
    >>> class Two(ABC):
    ...     pass
    >>> Two.__bases__ == (ABC, )
    True
    >>> Two.__bases__ == (Mixin, )
    False

- `mixin`:func: create a base-class tha consolidate several mix-ins and
  meta-classes.  For example::

    >>> from xoutil.eight.abc import ABCMeta

    >>> class One(dict):
    ...     pass

    >>> class Two(object):
    ...     pass

    >>> class OneMeta(type):
    ...     pass

    >>> class TwoMeta(type):
    ...     pass

    >>> Test = mixin(One, Two, meta=[OneMeta, TwoMeta, ABCMeta], name='Test')
    >>> Test.__name__ == 'Test'
    True
    >>> isinstance(Test, ABCMeta)
    True

These modules (this one and `meta`) must have four utilities:

- `metaclass` to use a unique syntax to declare meta-classes between Python 2
  and 3.

- `helper_class` to build a class that when used as a base impose a meta-class
  and not is found in resulting bases of defined class.  For example
  `xoutil.eight.abc.ABC`.

- `mixin` build a mixin-base composing several parts and meta-classes.

- `compose` specify the use of a mixin as one of the bases, but the new
  defined class will not be a mixin, this is not implemented yet because will
  require a big architecture re-factoring; for example::

    class Foobar(MyBase, compose(MyMixin)):
        pass

Maybe the last two names must be interchanged.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from xoutil.eight.meta import Mixin

from re import compile

_META_STRIP = compile('(?i)(^(meta(class)?|type)_{0,2}|'
                      '_{0,2}(meta(class)?|type)$)')

_MIXIN_STRIP = compile('(?i)(^mixin_{0,2}|_{0,2}mixin$)')


del compile


def helper_class(meta, name=None):
    '''Create a helper class based in the meta-class concept.

    :param meta: The meta-class type to base returned helper-class on it.

    :param name: The name (``__name__``) to assign to the returned class; if
           None is given, a nice name is calculated.

    For example::

      >>> from abc import ABCMeta
      >>> ABC = helper_class(ABCMeta)    # better than Python 3's abc.ABC :(
      >>> class MyError(Exception, ABC):
      ...     pass
      >>> (MyError.__bases__ == (Exception,), hasattr(MyError, 'register'))
      (True, True)

    This function calls `metaclass`:func: internally.  So, in the example
    anterior, `MyError` declaration is equivalent to::

      >>> class MyError(Exception, metaclass(ABCMeta)):
      ...     pass

    '''
    from xoutil.eight.meta import metaclass
    res = metaclass(meta)
    doc = ('Helper class.\n\nProvide a standard way to create classes with '
           'the meta-class `{meta}` using inheritance.\n\nFor example::\n\n'
           '  class My{name}({name}):\n      pass')
    if name and meta.__doc__:
        doc += ('\n\n`{meta}` is: ' + meta.__doc__)
    name = name or _META_STRIP.sub('', meta.__name__)
    res.__name__ = name
    res.__doc__ = (doc.strip() + '\n\n').format(name=name, meta=meta.__doc__)
    if meta.__module__:
        res.__module__ = meta.__module__
    return res


def mixin(*args, **kwargs):
    '''Weave a `mixin`.

    Parameters of this function are a little tricky.

    :param name: The name of the new class created by this function.  Could be
           passed as positional or keyword argument.  Use ``__name__`` as an
           alias.  See `helper_class`:func: for more info about this parameter
           and next two.

    :param doc: Documentation of the returned mixin-class.  Could be passed as
           positional or keyword argument.  Use ``__doc__`` as an alias.


    :param module: Always given as a keyword parameter.  A string -or an
           object to be used as reference- representing the module name.  Use
           ``__module__`` as an alias.

    :param metaclass: Always given as a keyword parameter.  Could be one type
           value or a list of values (multiples meta-classes).  Use
           (``__metaclass__``, ``metaclasses``, or ``meta``) as aliases.

    If several mixins with the same base are used all-together in a class
    inheritance, Python generates ``TypeError: multiple bases have instance
    lay-out conflict``.   To avoid that, inherit from the class this function
    returns instead of desired `base`.

    '''
    from xoutil.eight.meta import metaclass
    name, doc, bases = _name_doc_and_bases(args, kwargs)
    metas = _get_metaclasses(kwargs)
    mod = _get_module(bases, kwargs)
    mod = _calc_module(mod)
    _check_repeated(kwargs)
    if metas:
        kw = {'module': mod}
        if len(metas) == 1:
            meta = metas[0]
        else:
            meta = mixin(*metas, name='MultiMixinMeta', __nobase__=True, **kw)
        bases = bases + (metaclass(meta, **kw), )
    attrs = dict(kwargs, __doc__=doc)
    if mod:
        attrs['__module__'] = mod
    res = type(name, bases, attrs)
    # Mixin.register(type(res))
    return res


def _isname(s):
    '''Determine if `s` is a `mixin` name or not.'''
    from keyword import iskeyword
    from xoutil.eight import string_types
    from xoutil.eight.string import isidentifier
    if isinstance(s, string_types):
        res = isidentifier(s)
        if not (res and iskeyword(res)):
            return res
        else:
            msg = '''Reserved keyword "{}" can't be used as name'''
            raise ValueError(msg.format(res))
    else:
        return False


def _name_doc_and_bases(args, kwargs):
    '''Extract from positional args the name, doc and bases for a mixin.'''
    from xoutil.eight import string_types
    args = list(args)
    name = None
    doc = None
    i, count = 0, len(args)
    while i < count and not doc:
        s = args[i]
        if not name:
            aux = _isname(s)
            if aux:
                name = aux
                args.pop(i)
                count -= 1
            else:
                i += 1
        elif isinstance(s, string_types):
            s = str(s).strip()
            if s:
                doc = s
                args.pop(i)
                count -= 1
            else:
                raise ValueError('Invalid empty `mixin` documentation.')
        else:
            i += 1
    bases = _get_canonical_bases(args, kwargs)
    if not name:
        name = _get_kwarg(('__name__', 'name'), kwargs)
        if name:
            name = str(name)
        else:
            if bases:
                base = bases[0]
                aux = 'Basic' if base is Mixin else base.__name__
                aux = _MIXIN_STRIP.sub('', aux)
                suffix = '_mixin' if aux[0].islower() else 'Mixin'
                name = aux + suffix
            else:
                name = 'Mixin'
    if not doc:
        doc = _get_kwarg(('__doc__', 'doc'), kwargs)
        if not doc:
            names = ', '.join(b.__name__ for b in bases if b is not Mixin)
            if names:
                doc = 'Generated mixin base from ({}).'.format(names)
                aux = bases[0].__doc__
                if aux:
                    doc += ('\n\n' + aux)
            else:
                doc = 'Generated mixin base.'
    return name, doc, bases


def _get_kwarg(keys, kwargs):
    aux = (v for v in (kwargs.pop(k, None) for k in keys) if v is not None)
    return next(aux, None)


def _check_repeated(kwargs):
    keys = ('__name__', '__doc__', 'metaclass', '__metaclass__', '__module__')
    aux = {k for k, v in ((k, kwargs.get(k)) for k in keys) if v is not None}
    if aux:
        raise ValueError('Repeated keyword arguments: {}'.format(aux))


def _get_canonical_bases(args, kwargs):
    res = []
    abm = not kwargs.pop('__nobase__', False)
    for arg in args:
        i, count = 0, len(res)
        found = False
        while not found and i < count:
            if arg in type.mro(res[i]):    # `issubclass` not used because ABCs
                found = True
            elif res[i] in type.mro(arg):
                res[i] = arg
                found = True
            else:
                i += 1
        if not found:
            res.append(arg)
        if not abm and Mixin in type.mro(arg):
            abm = False
    if abm:
        res.append(Mixin)
    return tuple(res)


def _get_metaclasses(kwargs):
    names = ('metaclass', 'meta', 'metaclasses', '__metaclass__')
    res = _get_kwarg(names, kwargs)
    if res:
        res = tuple(res) if isinstance(res, (tuple, list)) else (res, )
    return res


def _get_module(bases, kwargs):
    from xoutil.eight import string_types
    res = _get_kwarg(('module', '__module__'), kwargs)
    if res:
        if isinstance(res, string_types):
            res = str(res)
        elif isinstance(res, bool):
            aux = bases[0].__module__
            res = aux if aux != object.__module__ else None
        else:
            res = res.__module__
    return res


def _get_frame_module():
    DEPTH = 3
    try:
        from sys import _getframe
    except ImportError:
        def _getframe(arg):
            attrs = dict(f_globals={'__name__': '<mixin>'})
            return type('frame', (object,), attrs)

    return _getframe(DEPTH).f_globals.get('__name__')


def _calc_module(mod):
    from xoutil.eight import string_types as strs
    _type = isinstance(mod, type)
    if mod or _type:
        if _type:
            return mod.__module__
        elif isinstance(mod, strs):
            mod = str(mod)
            return mod.format(_get_frame_module()) if '{}' in mod else mod
        else:
            return _get_frame_module()
    else:
        return None
